/* MoarNES source - NES.c

   This open-source software is (c)2010-2012 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This file contains the code to simulate the internal NES architecture,
   such as memory mapping, redirecting read/writes to I/O registers, etc.
*/

#include "config.h"
#include <stdio.h>
#include "common.h"
#include "stdint.h"
#include "rom.h"
#include "PPU.h"
#include "input.h"

/*
   Nintendo Entertainment System: CPU memory map

   $0000-$07FF = 2 KB internal RAM
   $0800-$1FFF = Mirrors of $0000-$07FF
   $2000-$2007 = PPU registers
   $2008-$3FFF = Mirrors of $2000-$2007
   $4000-$4017 = APU and I/O registers
   $4018-$FFFF = Cartridge PRG ROM, cartridge PRG RAM, and mapper registers

   $FFFA = NMI vector
   $FFFC = Reset vector
   $FFFE = IRQ/BRK vector

   In some cartridges, $6000-$7FFF is SRAM.

*/

uint32_t totalframes = 0;
uint8_t NESRAM[0x800]; //2 KB internal RAM
uint32_t cpugoal = 0, cpuadjust;
uint16_t sprtablesave;
uint8_t persistenttext[64];
uint16_t texttimer = 0;
uint16_t playspeed = 60;
struct cheat_s cheat[3];

extern uint8_t gamegenie;
extern void geniewrite(uint16_t addr, uint8_t value);

extern struct cart_s cartridge; //struct is declared in rom.c
extern struct nsf_s nsf;
extern bool ROMisNSF;
extern struct PPU_s *PPU;
extern struct OAM_s *OAM;

extern struct map4_s *map4;

extern uint64_t exec6502(uint64_t tickcount);
extern void step6502();
extern void reset6502();
extern void nmi6502();
extern uint16_t getPC();
extern uint64_t getclockticks();

extern uint8_t readPPU(uint16_t addr);
extern void writePPU(uint16_t addr, uint8_t value);
extern uint8_t readPPUregs(uint16_t addr);
extern void writePPUregs(uint16_t addr, uint8_t value);
extern void renderscanline(uint16_t scanline);

extern uint8_t processinput();
extern uint8_t joyread(uint8_t joynum);
extern void joystrobe();

extern uint8_t readAPU(uint16_t addr);
extern void writeAPU(uint16_t addr, uint8_t value);

extern uint8_t mapperwrite(uint16_t addr, uint8_t value);

extern uint8_t outputNES[240][256];
extern void blitframe();

extern void initrecord();

extern uint8_t savestatefilename[MAX_PATH];

#ifndef _WIN32
	struct timeval timestr;
#else
	LARGE_INTEGER dummytimer;
#endif
uint64_t timerfreq, starttimer, endtimer, difftimer, curtimer, nextframe, latestframe, frametime;
double doubletimer;
extern uint64_t nextsamplecycle, nextavisamplecycle, sampleticks, avisampleticks, nextseqcycle, seqtickgap;
extern uint8_t dorecord;
extern void drawtext(uint8_t *text, uint16_t x, uint16_t y);

extern void savestate();
extern void loadstate();

bool running = false;
bool paused = false;
bool doloadstate = false;
bool dosavestate = false;
bool doloadmovie = false;
bool dosavemovie = false;
bool playingmovie = false;
bool isquicksave = false;
uint32_t movieframe = 0, quicksavemovieframe[4], saveslot;

extern uint16_t pc;
extern uint8_t a, x;

#ifdef NETPLAY_ENABLE
	extern void sendudp(uint8_t *data, int16_t len);
	extern void processnet(uint32_t cursync);
#else
	void sendudp(uint8_t *data, int16_t len) {
	}
	void processnet(uint32_t cursync) {
	}
#endif
uint8_t lastsendbuf[12];
uint32_t netplaysyncgap = 3;
extern uint8_t netsynced;
uint8_t netmode;

uint8_t read6502(uint16_t addr) {
	uint8_t retval, cnum;

	if (addr < 0x2000) retval = NESRAM[addr & 0x7FF];
	else if (addr < 0x4000) retval = readPPUregs(0x2000 | (addr & 7));
	else if (addr < 0x4018) {
		switch (addr) {
			case 0x4016: if (netmode == 2) retval = joyread(1); else retval = joyread(0); break;
			case 0x4017: if (netmode == 2) retval = joyread(0); else retval = joyread(1); break;
			default: retval = readAPU(addr); break;
		}
	}
	else if (addr < 0x8000) {
		if (cartridge.hasSRAM == true) {
			retval = cartridge.SRAM[addr & 0x1FFF];
		} else retval = 0; //open data bus if no cartidge SRAM
	} else {
		//if we get to this point, it's a read from the cartridge
		addr &= 0x7FFF;
		retval = cartridge.PRGbank[addr>>10][addr&1023];
	}
	if (gamegenie == 2) {
		for (cnum=0; cnum<3; cnum++) {
			if ((cheat[cnum].valid) && (addr == cheat[cnum].addr)) if (cheat[cnum].docompare == 1) {
				if (retval == cheat[cnum].compare) retval = cheat[cnum].replace;
			} else {
				retval = cheat[cnum].replace;
			}
		}
	}

	return(retval);
}

void write6502(uint16_t addr, uint8_t value) {
	uint16_t tempsprite;
	if ((cartridge.mapper>0) && (addr>0x7FFF)) if (mapperwrite(addr, value)) return;
	if ((gamegenie == 1) && (addr>0x7FFF)) { geniewrite(addr, value); return; }
	if (addr < 0x2000) { NESRAM[addr & 0x7FF] = value; return; }
	if (addr < 0x4000) { writePPUregs(0x2000 | (addr & 7), value); return; }
	if (addr < 0x4018) {
		if (addr == 0x4014) {
			for (tempsprite=0; tempsprite<256; tempsprite++) {
				OAM->RAM[(OAM->addr+tempsprite)&255] = read6502(value*256+tempsprite);
			}
			return;
		}
		if (addr == 0x4016) joystrobe();
			else writeAPU(addr, value);
		return;
	}

	if (addr < 0x8000) {
		if (cartridge.hasSRAM == true) {
			cartridge.SRAM[addr & 0x1FFF] = value;
		} else return; //open data bus if no cartidge SRAM
	}

	//if we get to this point, it's an attempted write to the cartridge.
	//obviously, being ROM, nothing would actually get written anywhere.
}

extern void irq6502();
extern void map4irqdecrement();
void execframe(bool skiprender) {
	uint16_t scanline = 0;

	if (cartridge.mapper == 4) map4->irqcounter = map4->irqlatch;
	if (PPU->bgvisible) {
		if ((scanline%3)==0) exec6502(86);
			else exec6502(85);
		PPU->addr = PPU->tempaddr; //not sure if required, but doesn't seem to make a difference either way
		PPU->vblank = 0;
		PPU->sprzero = 0;
		PPU->sprover = 0;
		exec6502(28);
	} else exec6502(114);
	for (scanline=0; scanline<257; scanline++) {
		if ((scanline%3)==0) exec6502(86);
			else exec6502(85);

		if (scanline <240) {
			if (PPU->bgvisible) PPU->addr = (PPU->addr & 0xFBE0) | (PPU->tempaddr & 0x041F) | ((PPU->yscroll & 7) << 12);
			renderscanline(scanline);

			if (cartridge.mapper == 4) {
				if (PPU->bgvisible || PPU->sprvisible) {
					map4irqdecrement();
					if ((map4->irqenable) && (map4->irqcounter==0)) irq6502();
				}
			}
	
		}
		exec6502(28);

		if (scanline == 241) {
			PPU->vblank = 1;
			if (PPU->nmivblank) nmi6502();
		}
	}

	totalframes++;
}

uint8_t lol[32];
uint32_t recvframe = 0, sendframe = 0;
#ifdef NETPLAY_ENABLE
	extern uint8_t sendbuf[1024], recvbuf[1024];
#else
	uint8_t sendbuf[1024], recvbuf[1024];
#endif
extern uint8_t VRAM[16384];
extern struct inputdev_s gamepad[2];

extern int32_t avibufptr;
extern uint8_t avibuf[16384];
extern uint32_t avisamplesperframe;
extern void aviencodeframe(uint8_t *NESbuffer, uint8_t *audiobuffer, uint16_t samplecount);
extern SDL_AudioSpec audio;
FILE *moviefile = NULL;

uint8_t NSFframe[240][256];
extern uint32_t samplesperframe;

extern uint8_t buf[16384];
extern int32_t bufferpos, buffersize;
bool donextframe = false;

extern void killAPU();

void execNES() {
	uint8_t resend = 0, moviebyte, curbutton, cursamp, cursamp2, lastsamp = 0;
	uint16_t clearpos, yline;
	uint32_t starttick;
	uint16_t linear = 0; //for NSF visualizer

	reset6502();
	memset(&NESRAM[0], 0, 2048);

	getfreq(timerfreq);

	frametime = (uint64_t)((double)timerfreq / (double)playspeed);
	nextsamplecycle = getclockticks() + sampleticks;
	nextavisamplecycle = getclockticks() + avisampleticks;
	nextseqcycle = getclockticks() + seqtickgap;
	getclock(starttimer);
    nextframe = starttimer + frametime;
	sendframe = 0xFFFFFFFF;
	recvframe = 0xFFFFFFFF;
	totalframes = 0;
	movieframe = 0;
	if (doloadmovie) {
		moviefile = fopen(&savestatefilename[0], "rb");
		if (moviefile != NULL) playingmovie = true;
	}
	if (dosavemovie) {
		moviefile = fopen(&savestatefilename[0], "wb");
		if (moviefile != NULL) playingmovie = false;
	}
	if (ROMisNSF) {
		memset(&cartridge.SRAM[0], 0, 8192);
		memset(&NESRAM[0], 0, 2048);
		for (clearpos=0x4000; clearpos<=0x4013; clearpos++) {
			if (clearpos==0x4010) writeAPU(clearpos, 0x10);
				else writeAPU(clearpos, 0x00);
		}
		writeAPU(0x4015, 0x0F);
		a = 0; //song 1 always for now
		x = 0;
		pc = nsf.initaddr;
		exec6502(16667); //give the init code a frame's worth of CPU ticks to finish
		memset(&NSFframe[0][0], 64, sizeof(NSFframe));
	}
	memset(&buf[0], 128, 16384);
	memset(&avibuf[0], 128, 16384);
	bufferpos = buffersize;
	avibufptr = buffersize;
	while (running) {
		frametime = (uint64_t)((double)timerfreq / (double)playspeed);
		if (playingmovie && (moviefile != NULL)) {
			processinput();
			fseek(moviefile, movieframe++, SEEK_SET);
			if (fread(&moviebyte, 1, 1, moviefile)) {
				for (curbutton=0; curbutton<8; curbutton++) {
					gamepad[0].gamepad.buttonvalue[curbutton] = (moviebyte >> curbutton) & 1;
				}
			} else {
				fclose(moviefile);
				playingmovie = false;
			}
		} else if (netmode == 0) {
			processinput();
			if (dosavemovie) {
				fseek(moviefile, movieframe++, SEEK_SET);
				moviebyte = 0;
				for (curbutton=0; curbutton<8; curbutton++) {
					moviebyte |= gamepad[0].gamepad.buttonvalue[curbutton] << curbutton;
				}
				fwrite(&moviebyte, 1, 1, moviefile);
			}
		} else {
			if ((totalframes % netplaysyncgap) == 0) {
				processinput();
				memcpy(&sendbuf[0], &sendframe, 4);
				memcpy(&sendbuf[4], &gamepad[0].gamepad.buttonvalue[0], 8);
				//memcpy(&savegamepad[0], &gamepad[0].gamepad.buttonvalue[0], 8);
				sendudp(&sendbuf[0], 12);
				netsynced = 0;
				starttick = SDL_GetTicks();
				while (1) {
					processnet(recvframe);
					if ((SDL_GetTicks() - starttick) > 250) {
						starttick = SDL_GetTicks();
						processinput();
					}
					//memcpy(&gamepad[0].gamepad.buttonvalue[0], &savegamepad[0], 8);
					if (!memcmp(&recvbuf[0], &recvframe, 4)) {
						memcpy(&gamepad[1].gamepad.buttonvalue[0], &recvbuf[4], 8);
						sprintf(&persistenttext[0], "Remote sync: %lu", recvframe);
						texttimer = 240;
						break;
					}
					if (!running) break;
					SDL_Delay(1);
					if (resend == 10) {
						sendudp(&sendbuf[0], 12);
					} else if (resend == 30) {
						sendudp(&lastsendbuf[0], 12);
					}
					resend = (resend + 1) % 40;
				}
				sendframe++;
				recvframe++;
				memcpy(&lastsendbuf[0], &sendbuf[0], 12);
				netsynced = 1;
			}
		}

		avibufptr = 0;
		if (ROMisNSF) {
			pc = nsf.playaddr;
			exec6502(16667);
			//sprintf(&persistenttext[0], "%04X %04X ", nsf.initaddr, nsf.playaddr);
		}
		execframe(false);
		if (ROMisNSF) {
			memset(&outputNES[0][0], 65, 61440);
			for (linear=0; linear<samplesperframe; linear++) {
				cursamp = avibuf[ (uint16_t)(((double)avibufptr / (double)samplesperframe) * (double)linear) ];
				if (cursamp<240) {
					if (cursamp>127) cursamp2 = ((~cursamp) & 127) / 43;
						else cursamp2 = (cursamp & 127) / 43;
					cursamp2 = cursamp2*16+10;
					//outputNES[cursamp][linear>>2] = cursamp2; //[(uint16_t)((double)linear / ((double)samplesperframe/(double)256.0))] = cursamp2;
					if (cursamp>lastsamp) {
						for (yline=lastsamp; yline<=cursamp; yline++) {
							outputNES[yline][linear] = cursamp2;
						}
					} else {
						for (yline=cursamp; yline<=lastsamp; yline++) {
							outputNES[yline][linear] = cursamp2;
						}
					}
					lastsamp = cursamp;
				}
			}
		}

		if (texttimer) {
			drawtext(&persistenttext[0], 0, 232);
			texttimer--;
		}

		while (bufferpos < buffersize) step6502();
		if (dorecord) aviencodeframe(&outputNES[0][0], &avibuf[0], (uint16_t)avibufptr);
		while (donextframe == false) {
			SDL_Delay(1);
		}
		donextframe = false;
		if ((curtimer - nextframe) > (frametime << 6)) nextframe = curtimer + frametime;
		nextframe += frametime;
		blitframe(&outputNES[0][0]);

		while (paused == true) {
			processinput();
			SDL_Delay(10);
		}
		if (dosavestate == true) {
			savestate();
			if (isquicksave) {
				isquicksave = false;
				quicksavemovieframe[saveslot] = movieframe;
			}
		}
		if (doloadstate == true) {
			loadstate();
			if (isquicksave) {
				isquicksave = false;
				movieframe = quicksavemovieframe[saveslot];
			}
		}
	}
	if (doloadmovie || dosavemovie) {
		if (moviefile != NULL) fclose(moviefile);
		doloadmovie = false;
		dosavemovie = false;
	}

	getclock(endtimer);
	killAPU();

	/*difftimer = ((endtimer * 1000) - (starttimer * 1000)) / timerfreq;
	doubletimer = (double)difftimer / (double)1000.0;
	printf("\n");
    printf("Executed %lu CPU clock ticks in %f seconds. (approx. %lu clock ticks/sec)\n", getclockticks(), doubletimer, (uint32_t)((double)getclockticks() / doubletimer));
	printf("Rendered %lu frames.\n", totalframes);*/
}

void timedtext(uint8_t *text) {
	sprintf(&persistenttext[0], text);
	texttimer = 240; //unit of measure is frames (at 60 FPS)
}
